use crate::internal::modbus::{BaseModbusCode, CsvData};
use anyhow::anyhow;
use serde_json::{Number, Value};

/// 寄存器类型
#[derive(Debug, Eq, PartialEq)]
pub enum RegisterType {
    //线圈寄存器 0
    Coil,
    //离散输入寄存器 1
    DiscreteInput,
    //保持寄存器 2
    HoldingRegister,
    //输入寄存器 3
    InputRegister,
}

impl TryFrom<u8> for RegisterType {
    type Error = anyhow::Error;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(RegisterType::Coil),
            1 => Ok(RegisterType::DiscreteInput),
            2 => Ok(RegisterType::HoldingRegister),
            3 => Ok(RegisterType::InputRegister),
            _ => anyhow::bail!("无效的寄存器类型！"),
        }
    }
}

/// 数据类型
#[derive(Debug)]
pub enum ModbusDataType {
    Bit,
    U16,
    I16,
}

impl TryFrom<String> for ModbusDataType {
    type Error = anyhow::Error;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        match value.as_str() {
            "bit" => Ok(ModbusDataType::Bit),
            "u16" => Ok(ModbusDataType::U16),
            "i16" => Ok(ModbusDataType::I16),
            _ => anyhow::bail!("无效的数据类型！"),
        }
    }
}

#[derive(Debug, Eq, PartialEq)]
pub enum Rw {
    R,
    W,
    RW,
}

impl TryFrom<String> for Rw {
    type Error = anyhow::Error;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        match value.as_str() {
            "r" => Ok(Rw::R),
            "w" => Ok(Rw::W),
            "rw" => Ok(Rw::RW),
            _ => anyhow::bail!("无效的读写！"),
        }
    }
}

#[derive(Debug)]
pub struct ModbusCode {
    pub addr: u16,
    pub code: RegisterType,
    pub data: Option<u16>,
    pub name: String,
    pub data_type: Option<ModbusDataType>,
    pub factor: Option<f32>,
    pub rw: Option<Rw>,
    pub offset: Option<i16>,
    pub key: Option<String>,
}

impl ModbusCode {

    /// 计算rawValue = data * factor + offset
    pub fn json_value(&self) -> Option<Value> {
        if self.data_type.is_some() && self.data.is_some() {
            let data_type = self.data_type.as_ref()?;
            let data = *self.data.as_ref()? as f32;
            let factor = *self.factor.as_ref()?;
            //真值 = 通讯数据 * 分辨率 + 偏移量
            let mut raw_data = data * factor;
            // 处理 offset（如果有）
            if let Some(offset) = self.offset.as_ref() {
                raw_data += *offset as f32;
            }
            // 计算最终值并四舍五入
            let raw_data = (raw_data / factor).round() * factor;
            match data_type {
                ModbusDataType::Bit => Some(Value::Number(Number::from(data as u8))),
                _ => {
                    if factor == 1f32 {
                        // factor=1 时直接返回整数（不带小数位）
                        Some(Value::Number(Number::from(raw_data as i64)))
                    } else {
                        // 计算小数位数（如 factor=0.1 → 1位小数）
                        let decimal_places = (-factor.log10()).round() as usize;
                        // 格式化字符串，确保正确的小数位数
                        let formatted = format!("{0:.1$}", raw_data, decimal_places);
                        // 解析回 f64 并转为 Number
                        let number = formatted.parse::<f64>().ok().and_then(Number::from_f64)?;
                        Some(Value::Number(number))
                    }
                }
            }
        } else {
            None
        }
    }
}

impl BaseModbusCode for ModbusCode {
    fn addr(&self) -> u16 {
        self.addr
    }
}

impl BaseModbusCode for &mut ModbusCode {
    fn addr(&self) -> u16 {
        self.addr
    }
}

impl TryFrom<CsvData> for ModbusCode {
    type Error = anyhow::Error;

    fn try_from(value: CsvData) -> Result<Self, Self::Error> {
        let addr = value.address.ok_or(anyhow!("寄存器地址为空！"))?;
        let code = value.code.ok_or(anyhow!("寄存器类型为空！"))?.try_into()?;
        let name = value.name.ok_or(anyhow!("寄存器名称为空！"))?;
        let data_type = match value.t {
            None => None,
            Some(it) => Some(it.try_into()?),
        };
        let rw = match value.rw {
            None => None,
            Some(it) => Some(it.try_into()?),
        };
        let offset = value.offset;
        let factor = value.factor;
        let key = value.key;
        Ok(ModbusCode {
            addr,
            code,
            data: None,
            name,
            data_type,
            factor,
            rw,
            offset,
            key,
        })
    }
}
